@@ -1,26 +1,38 @@
-# Copyright (c) 2001-2005 Twisted Matrix Laboratories.
+# Copyright (c) 2001-2009 Twisted Matrix Laboratories.
 # See LICENSE for details.
 
 """
-Test cases for twisted.words.protocols.msn
+Test cases for L{twisted.words.protocols.msn}.
 """
 
+# System imports
+import StringIO
+
 # Twisted imports
-from twisted.words.protocols import msn
+
+# t.w.p.msn requires an HTTP client
+try:
+    # So try to get one - do it directly instead of catching an ImportError
+    # from t.w.p.msn so that other problems which cause that module to fail
+    # to import don't cause the tests to be skipped.
+    from twisted.web import client
+except ImportError:
+    # If there isn't one, we're going to skip all the tests.
+    msn = None
+else:
+    # Otherwise importing it should work, so do it.
+    from twisted.words.protocols import msn
+
+
+from twisted.python.hashlib import md5
 from twisted.protocols import loopback
 from twisted.internet.defer import Deferred
 from twisted.trial import unittest
-
-# System imports
-import StringIO, sys
+from twisted.test.proto_helpers import StringTransport, StringIOWithoutClosing
 
 def printError(f):
     print f
 
-class StringIOWithoutClosing(StringIO.StringIO):
-    disconnecting = 0
-    def close(self): pass
-    def loseConnection(self): pass
 
 class PassportTests(unittest.TestCase):
 
@@ -30,23 +42,30 @@ class PassportTests(unittest.TestCase):
         self.deferred.addCallback(lambda r: self.result.append(r))
         self.deferred.addErrback(printError)
 
-    def testNexus(self):
+    def test_nexus(self):
+        """
+        When L{msn.PassportNexus} receives enough information to identify the
+        address of the login server, it fires the L{Deferred} passed to its
+        initializer with that address.
+        """
         protocol = msn.PassportNexus(self.deferred, 'https://foobar.com/somepage.quux')
         headers = {
             'Content-Length' : '0',
             'Content-Type'   : 'text/html',
             'PassportURLs'   : 'DARealm=Passport.Net,DALogin=login.myserver.com/,DAReg=reg.myserver.com'
         }
-        transport = StringIOWithoutClosing()
+        transport = StringTransport()
         protocol.makeConnection(transport)
         protocol.dataReceived('HTTP/1.0 200 OK\r\n')
-        for (h,v) in headers.items(): protocol.dataReceived('%s: %s\r\n' % (h,v))
+        for (h, v) in headers.items():
+            protocol.dataReceived('%s: %s\r\n' % (h,v))
         protocol.dataReceived('\r\n')
-        self.failUnless(self.result[0] == "https://login.myserver.com/")
+        self.assertEquals(self.result[0], "https://login.myserver.com/")
+
 
     def _doLoginTest(self, response, headers):
         protocol = msn.PassportLogin(self.deferred,'foo@foo.com','testpass','https://foo.com/', 'a')
-        protocol.makeConnection(StringIOWithoutClosing())
+        protocol.makeConnection(StringTransport())
         protocol.dataReceived(response)
         for (h,v) in headers.items(): protocol.dataReceived('%s: %s\r\n' % (h,v))
         protocol.dataReceived('\r\n')
@@ -82,63 +101,111 @@ class PassportTests(unittest.TestCase):
         self.failUnless(self.result[0] == (msn.LOGIN_REDIRECT, 'https://newlogin.host.com/', 'a'))
 
 
-class DummySwitchboardClient(msn.SwitchboardClient):
-    def userTyping(self, message):
-        self.state = 'TYPING'
+if msn is not None:
+    class DummySwitchboardClient(msn.SwitchboardClient):
+        def userTyping(self, message):
+            self.state = 'TYPING'
 
-    def gotSendRequest(self, fileName, fileSize, cookie, message):
-        if fileName == 'foobar.ext' and fileSize == 31337 and cookie == 1234: self.state = 'INVITATION'
+        def gotSendRequest(self, fileName, fileSize, cookie, message):
+            if fileName == 'foobar.ext' and fileSize == 31337 and cookie == 1234: self.state = 'INVITATION'
 
 
-class DummyNotificationClient(msn.NotificationClient):
-    def loggedIn(self, userHandle, screenName, verified):
-        if userHandle == 'foo@bar.com' and screenName == 'Test Screen Name' and verified:
-            self.state = 'LOGIN'
+    class DummyNotificationClient(msn.NotificationClient):
+        def loggedIn(self, userHandle, screenName, verified):
+            if userHandle == 'foo@bar.com' and screenName == 'Test Screen Name' and verified:
+                self.state = 'LOGIN'
 
-    def gotProfile(self, message):
-        self.state = 'PROFILE'
+        def gotProfile(self, message):
+            self.state = 'PROFILE'
 
-    def gotContactStatus(self, code, userHandle, screenName):
-        if code == msn.STATUS_AWAY and userHandle == "foo@bar.com" and screenName == "Test Screen Name":
-            self.state = 'INITSTATUS'
+        def gotContactStatus(self, code, userHandle, screenName):
+            if code == msn.STATUS_AWAY and userHandle == "foo@bar.com" and screenName == "Test Screen Name":
+                self.state = 'INITSTATUS'
 
-    def contactStatusChanged(self, code, userHandle, screenName):
-        if code == msn.STATUS_LUNCH and userHandle == "foo@bar.com" and screenName == "Test Name":
-            self.state = 'NEWSTATUS'
+        def contactStatusChanged(self, code, userHandle, screenName):
+            if code == msn.STATUS_LUNCH and userHandle == "foo@bar.com" and screenName == "Test Name":
+                self.state = 'NEWSTATUS'
 
-    def contactOffline(self, userHandle):
-        if userHandle == "foo@bar.com": self.state = 'OFFLINE'
+        def contactOffline(self, userHandle):
+            if userHandle == "foo@bar.com": self.state = 'OFFLINE'
 
-    def statusChanged(self, code):
-        if code == msn.STATUS_HIDDEN: self.state = 'MYSTATUS'
+        def statusChanged(self, code):
+            if code == msn.STATUS_HIDDEN: self.state = 'MYSTATUS'
 
-    def listSynchronized(self, *args):
-        self.state = 'GOTLIST'
+        def listSynchronized(self, *args):
+            self.state = 'GOTLIST'
 
-    def gotPhoneNumber(self, listVersion, userHandle, phoneType, number):
-        msn.NotificationClient.gotPhoneNumber(self, listVersion, userHandle, phoneType, number)
-        self.state = 'GOTPHONE'
+        def gotPhoneNumber(self, listVersion, userHandle, phoneType, number):
+            msn.NotificationClient.gotPhoneNumber(self, listVersion, userHandle, phoneType, number)
+            self.state = 'GOTPHONE'
 
-    def userRemovedMe(self, userHandle, listVersion):
-        msn.NotificationClient.userRemovedMe(self, userHandle, listVersion)
-        c = self.factory.contacts.getContact(userHandle)
-        if not c and self.factory.contacts.version == listVersion: self.state = 'USERREMOVEDME'
+        def userRemovedMe(self, userHandle, listVersion):
+            msn.NotificationClient.userRemovedMe(self, userHandle, listVersion)
+            c = self.factory.contacts.getContact(userHandle)
+            if not c and self.factory.contacts.version == listVersion: self.state = 'USERREMOVEDME'
+
+        def userAddedMe(self, userHandle, screenName, listVersion):
+            msn.NotificationClient.userAddedMe(self, userHandle, screenName, listVersion)
+            c = self.factory.contacts.getContact(userHandle)
+            if c and (c.lists | msn.REVERSE_LIST) and (self.factory.contacts.version == listVersion) and \
+               (screenName == 'Screen Name'):
+                self.state = 'USERADDEDME'
+
+        def gotSwitchboardInvitation(self, sessionID, host, port, key, userHandle, screenName):
+            if sessionID == 1234 and \
+               host == '192.168.1.1' and \
+               port == 1863 and \
+               key == '123.456' and \
+               userHandle == 'foo@foo.com' and \
+               screenName == 'Screen Name':
+                self.state = 'SBINVITED'
+
+
+
+class DispatchTests(unittest.TestCase):
+    """
+    Tests for L{DispatchClient}.
+    """
+    def _versionTest(self, serverVersionResponse):
+        """
+        Test L{DispatchClient} version negotiation.
+        """
+        client = msn.DispatchClient()
+        client.userHandle = "foo"
+
+        transport = StringTransport()
+        client.makeConnection(transport)
+        self.assertEquals(
+            transport.value(), "VER 1 MSNP8 CVR0\r\n")
+        transport.clear()
+
+        client.dataReceived(serverVersionResponse)
+        self.assertEquals(
+            transport.value(),
+            "CVR 2 0x0409 win 4.10 i386 MSNMSGR 5.0.0544 MSMSGS foo\r\n")
+
+
+    def test_version(self):
+        """
+        L{DispatchClient.connectionMade} greets the server with a I{VER}
+        (version) message and then L{NotificationClient.dataReceived}
+        handles the server's I{VER} response by sending a I{CVR} (client
+        version) message.
+        """
+        self._versionTest("VER 1 MSNP8 CVR0\r\n")
+
+
+    def test_versionWithoutCVR0(self):
+        """
+        If the server responds to a I{VER} command without including the
+        I{CVR0} protocol, L{DispatchClient} behaves in the same way as if
+        that protocol were included.
+
+        Starting in August 2008, CVR0 disappeared from the I{VER} response.
+        """
+        self._versionTest("VER 1 MSNP8\r\n")
 
-    def userAddedMe(self, userHandle, screenName, listVersion):
-        msn.NotificationClient.userAddedMe(self, userHandle, screenName, listVersion)
-        c = self.factory.contacts.getContact(userHandle)
-        if c and (c.lists | msn.REVERSE_LIST) and (self.factory.contacts.version == listVersion) and \
-           (screenName == 'Screen Name'):
-            self.state = 'USERADDEDME'
 
-    def gotSwitchboardInvitation(self, sessionID, host, port, key, userHandle, screenName):
-        if sessionID == 1234 and \
-           host == '192.168.1.1' and \
-           port == 1863 and \
-           key == '123.456' and \
-           userHandle == 'foo@foo.com' and \
-           screenName == 'Screen Name':
-            self.state = 'SBINVITED'
 
 class NotificationTests(unittest.TestCase):
     """ testing the various events in NotificationClient """
@@ -148,13 +215,78 @@ class NotificationTests(unittest.TestCase):
         self.client.factory = msn.NotificationFactory()
         self.client.state = 'START'
 
+
     def tearDown(self):
         self.client = None
 
+
+    def _versionTest(self, serverVersionResponse):
+        """
+        Test L{NotificationClient} version negotiation.
+        """
+        self.client.factory.userHandle = "foo"
+
+        transport = StringTransport()
+        self.client.makeConnection(transport)
+        self.assertEquals(
+            transport.value(), "VER 1 MSNP8 CVR0\r\n")
+        transport.clear()
+
+        self.client.dataReceived(serverVersionResponse)
+        self.assertEquals(
+            transport.value(),
+            "CVR 2 0x0409 win 4.10 i386 MSNMSGR 5.0.0544 MSMSGS foo\r\n")
+
+
+    def test_version(self):
+        """
+        L{NotificationClient.connectionMade} greets the server with a I{VER}
+        (version) message and then L{NotificationClient.dataReceived}
+        handles the server's I{VER} response by sending a I{CVR} (client
+        version) message.
+        """
+        self._versionTest("VER 1 MSNP8 CVR0\r\n")
+
+
+    def test_versionWithoutCVR0(self):
+        """
+        If the server responds to a I{VER} command without including the
+        I{CVR0} protocol, L{NotificationClient} behaves in the same way as
+        if that protocol were included.
+
+        Starting in August 2008, CVR0 disappeared from the I{VER} response.
+        """
+        self._versionTest("VER 1 MSNP8\r\n")
+
+
+    def test_challenge(self):
+        """
+        L{NotificationClient} responds to a I{CHL} message by sending a I{QRY}
+        back which included a hash based on the parameters of the I{CHL}.
+        """
+        transport = StringTransport()
+        self.client.makeConnection(transport)
+        transport.clear()
+
+        challenge = "15570131571988941333"
+        self.client.dataReceived('CHL 0 ' + challenge + '\r\n')
+        # md5 of the challenge and a magic string defined by the protocol
+        response = "8f2f5a91b72102cd28355e9fc9000d6e"
+        # Sanity check - the response is what the comment above says it is.
+        self.assertEquals(
+            response, md5(challenge + "Q1P7W2E4J9R8U3S5").hexdigest())
+        self.assertEquals(
+            transport.value(),
+            # 2 is the next transaction identifier.  32 is the length of the
+            # response.
+            "QRY 2 msmsgs@msnmsgr.com 32\r\n" + response)
+
+
     def testLogin(self):
         self.client.lineReceived('USR 1 OK foo@bar.com Test%20Screen%20Name 1 0')
         self.failUnless((self.client.state == 'LOGIN'), msg='Failed to detect successful login')
 
+
     def testProfile(self):
         m = 'MSG Hotmail Hotmail 353\r\nMIME-Version: 1.0\r\nContent-Type: text/x-msmsgsprofile; charset=UTF-8\r\n'
         m += 'LoginTime: 1016941010\r\nEmailEnabled: 1\r\nMemberIdHigh: 40000\r\nMemberIdLow: -600000000\r\nlang_preference: 1033\r\n'
@@ -177,7 +309,7 @@ class NotificationTests(unittest.TestCase):
         # that BPRs sent as part of the SYN reply may not be interpreted
         # as such if they are for the last LST -- maybe I should
         # factor this in later.
-        self.client.makeConnection(StringIOWithoutClosing())
+        self.client.makeConnection(StringTransport())
         msn.NotificationClient.loggedIn(self.client, 'foo@foo.com', 'foobar', 1)
         lines = [
             "SYN %s 100 1 1" % self.client.currentID,
@@ -199,7 +331,7 @@ class NotificationTests(unittest.TestCase):
         c = msn.MSNContact(userHandle='userHandle@email.com')
         self.client.factory.contacts = msn.MSNContactList()
         self.client.factory.contacts.addContact(c)
-        self.client.makeConnection(StringIOWithoutClosing())
+        self.client.makeConnection(StringTransport())
         self.client.lineReceived("BPR 101 userHandle@email.com PHH 123%20456")
         c = self.client.factory.contacts.getContact('userHandle@email.com')
         self.failUnless(self.client.state == 'GOTPHONE', "Did not fire phone change callback")
@@ -212,7 +344,7 @@ class NotificationTests(unittest.TestCase):
         to be part of a SYN response (but came after the last LST)
         is received, the correct contact is updated and all is well
         """
-        self.client.makeConnection(StringIOWithoutClosing())
+        self.client.makeConnection(StringTransport())
         msn.NotificationClient.loggedIn(self.client, 'foo@foo.com', 'foo', 1)
         lines = [
             "SYN %s 100 1 1" % self.client.currentID,
@@ -243,6 +375,19 @@ class NotificationTests(unittest.TestCase):
         self.client.lineReceived("RNG 1234 192.168.1.1:1863 CKI 123.456 foo@foo.com Screen%20Name")
         self.failUnless(self.client.state == "SBINVITED")
 
+    def testCommandFailed(self):
+        """
+        Ensures that error responses from the server fires an errback with
+        MSNCommandFailed.
+        """
+        id, d = self.client._createIDMapping()
+        self.client.lineReceived("201 %s" % id)
+        d = self.assertFailure(d, msn.MSNCommandFailed)
+        def assertErrorCode(exception):
+            self.assertEqual(201, exception.errorCode)
+        return d.addCallback(assertErrorCode)
+
+
 class MessageHandlingTests(unittest.TestCase):
     """ testing various message handling methods from SwichboardClient """
 
@@ -314,25 +459,45 @@ class MessageHandlingTests(unittest.TestCase):
 
 
 class FileTransferTestCase(unittest.TestCase):
-    """ test FileSend against FileReceive """
+    """
+    test FileSend against FileReceive
+    """
 
     def setUp(self):
-        self.input = StringIOWithoutClosing()
-        self.input.writelines(['a'] * 7000)
-        self.input.seek(0)
+        self.input = 'a' * 7000
         self.output = StringIOWithoutClosing()
 
+
     def tearDown(self):
         self.input = None
         self.output = None
 
-    def testFileTransfer(self):
+
+    def test_fileTransfer(self):
+        """
+        Test L{FileSend} against L{FileReceive} using a loopback transport.
+        """
         auth = 1234
-        sender = msn.FileSend(self.input)
+        sender = msn.FileSend(StringIO.StringIO(self.input))
         sender.auth = auth
         sender.fileSize = 7000
         client = msn.FileReceive(auth, "foo@bar.com", self.output)
         client.fileSize = 7000
-        loopback.loopback(sender, client)
-        self.failUnless((client.completed and sender.completed), msg="send failed to complete")
-        self.failUnless((self.input.getvalue() == self.output.getvalue()), msg="saved file does not match original")
+        def check(ignored):
+            self.assertTrue(
+                client.completed and sender.completed,
+                msg="send failed to complete")
+            self.assertEqual(
+                self.input, self.output.getvalue(),
+                msg="saved file does not match original")
+        d = loopback.loopbackAsync(sender, client)
+        d.addCallback(check)
+        return d
+
+
+if msn is None:
+    for testClass in [PassportTests, NotificationTests,
+                      MessageHandlingTests, FileTransferTestCase]:
+        testClass.skip = (
+            "MSN requires an HTTP client but none is available, "
+            "skipping tests.")
